pub mod algo;
mod generated;

pub use generated::*;
use itertools::Itertools;
use std::{fmt::Debug, marker::PhantomData};
use syntax::{SyntaxElement, SyntaxKind::*, TomlVersion, T};

pub trait AstNode
where
    Self: Debug,
{
    fn leading_comments(&self) -> impl Iterator<Item = crate::Comment> {
        support::leading_comments(self.syntax().children_with_tokens())
    }

    fn tailing_comment(&self) -> Option<crate::Comment> {
        self.syntax().last_token().and_then(crate::Comment::cast)
    }

    fn can_cast(kind: syntax::SyntaxKind) -> bool
    where
        Self: Sized;

    fn cast(syntax: syntax::SyntaxNode) -> Option<Self>
    where
        Self: Sized;

    fn syntax(&self) -> &syntax::SyntaxNode;
}

/// Like `AstNode`, but wraps tokens rather than interior nodes.
pub trait AstToken {
    fn can_cast(token: syntax::SyntaxKind) -> bool
    where
        Self: Sized;

    fn cast(syntax: syntax::SyntaxToken) -> Option<Self>
    where
        Self: Sized;

    fn syntax(&self) -> &syntax::SyntaxToken;

    fn text(&self) -> &str {
        self.syntax().text()
    }
}

#[allow(dead_code)]
#[derive(Debug, Clone)]
pub struct AstChildren<N> {
    inner: syntax::SyntaxNodeChildren,
    ph: PhantomData<N>,
}

impl<N> AstChildren<N> {
    fn new(parent: &syntax::SyntaxNode) -> Self {
        AstChildren {
            inner: parent.children(),
            ph: PhantomData,
        }
    }
}

impl<N: AstNode> Iterator for AstChildren<N> {
    type Item = N;
    fn next(&mut self) -> Option<N> {
        self.inner.find_map(N::cast)
    }
}

#[allow(dead_code)]
mod support {
    use syntax::SyntaxKind;

    use super::*;

    #[inline]
    pub(super) fn child<N: AstNode>(parent: &syntax::SyntaxNode) -> Option<N> {
        parent.children().find_map(N::cast)
    }

    #[inline]
    pub(super) fn children<N: AstNode>(parent: &syntax::SyntaxNode) -> AstChildren<N> {
        AstChildren::new(parent)
    }

    #[inline]
    pub(super) fn token(
        parent: &syntax::SyntaxNode,
        kind: syntax::SyntaxKind,
    ) -> Option<syntax::SyntaxToken> {
        parent
            .children_with_tokens()
            .filter_map(|it| it.into_token())
            .find(|it| it.kind() == kind)
    }

    #[inline]
    pub(super) fn leading_comments<I: Iterator<Item = syntax::SyntaxElement>>(
        iter: I,
    ) -> impl Iterator<Item = crate::Comment> {
        iter.take_while(|node| matches!(node.kind(), COMMENT | LINE_BREAK | WHITESPACE))
            .filter_map(|node_or_token| match node_or_token {
                SyntaxElement::Token(token) => crate::Comment::cast(token),
                SyntaxElement::Node(_) => None,
            })
    }

    #[inline]
    pub(super) fn tailing_comment<I: Iterator<Item = syntax::SyntaxElement>>(
        iter: I,
        end: syntax::SyntaxKind,
    ) -> Option<crate::Comment> {
        let mut iter = iter
            .skip_while(|item| item.kind() != end && item.kind() != EOF)
            .skip(1);

        match iter.next()? {
            SyntaxElement::Token(token) if token.kind() == COMMENT => crate::Comment::cast(token),
            SyntaxElement::Token(token) if token.kind() == WHITESPACE => {
                iter.next().and_then(|node_or_token| match node_or_token {
                    SyntaxElement::Token(token) if token.kind() == COMMENT => {
                        crate::Comment::cast(token)
                    }
                    _ => None,
                })
            }
            _ => None,
        }
    }

    #[inline]
    pub(super) fn begin_dangling_comments<I: Iterator<Item = syntax::SyntaxElement>>(
        iter: I,
    ) -> impl Iterator<Item = crate::Comment> {
        iter.take_while(|node| matches!(node.kind(), COMMENT | WHITESPACE | LINE_BREAK))
            .filter_map(|node_or_token| match node_or_token {
                SyntaxElement::Token(token) => crate::Comment::cast(token),
                SyntaxElement::Node(_) => None,
            })
    }

    #[inline]
    pub(super) fn end_dangling_comments<I: Iterator<Item = syntax::SyntaxElement>>(
        iter: I,
    ) -> impl Iterator<Item = crate::Comment> {
        iter.collect::<Vec<_>>()
            .into_iter()
            .rev()
            .take_while(|node| matches!(node.kind(), COMMENT | WHITESPACE | LINE_BREAK))
            .filter_map(|node_or_token| match node_or_token {
                SyntaxElement::Token(token) => crate::Comment::cast(token),
                SyntaxElement::Node(_) => None,
            })
            .collect::<Vec<_>>()
            .into_iter()
            .rev()
    }

    #[inline]
    pub(super) fn has_inner_comments<I: Iterator<Item = syntax::SyntaxElement>>(
        iter: I,
        start: SyntaxKind,
        end: SyntaxKind,
    ) -> bool {
        iter.skip_while(|node| node.kind() != start)
            .skip(1)
            .take_while(|node| node.kind() != end)
            .any(|node| {
                node.kind() == COMMENT
                    || match node {
                        syntax::SyntaxElement::Node(node) => node
                            .children_with_tokens()
                            .any(|node| node.kind() == COMMENT),
                        _ => false,
                    }
            })
    }
}

impl Root {
    pub fn begin_dangling_comments(&self) -> impl Iterator<Item = crate::Comment> {
        support::begin_dangling_comments(self.syntax().children_with_tokens())
    }

    pub fn end_dangling_comments(&self) -> impl Iterator<Item = crate::Comment> {
        support::end_dangling_comments(self.syntax().children_with_tokens())
    }
}

impl Table {
    pub fn header_leading_comments(&self) -> impl Iterator<Item = crate::Comment> {
        support::leading_comments(self.syntax().children_with_tokens())
    }

    pub fn header_tailing_comment(&self) -> Option<crate::Comment> {
        support::tailing_comment(self.syntax().children_with_tokens(), T!(']'))
    }
}

impl ArrayOfTable {
    pub fn header_leading_comments(&self) -> impl Iterator<Item = crate::Comment> {
        support::leading_comments(self.syntax().children_with_tokens())
    }

    pub fn header_tailing_comment(&self) -> Option<crate::Comment> {
        support::tailing_comment(self.syntax().children_with_tokens(), T!("]]"))
    }
}

impl Array {
    pub fn inner_begin_dangling_comments(&self) -> impl Iterator<Item = crate::Comment> {
        support::begin_dangling_comments(
            self.syntax()
                .children_with_tokens()
                .skip_while(|node| node.kind() != T!('['))
                .skip(1),
        )
    }

    pub fn inner_end_dangling_comments(&self) -> impl Iterator<Item = crate::Comment> {
        support::end_dangling_comments(
            self.syntax()
                .children_with_tokens()
                .take_while(|node| node.kind() != T!(']')),
        )
    }

    pub fn values_with_comma(&self) -> impl Iterator<Item = (crate::Value, Option<crate::Comma>)> {
        self.values()
            .zip_longest(support::children::<crate::Comma>(self.syntax()))
            .map(|value_with_comma| match value_with_comma {
                itertools::EitherOrBoth::Both(value, comma) => (value, Some(comma)),
                itertools::EitherOrBoth::Left(value) => (value, None),
                itertools::EitherOrBoth::Right(_) => unreachable!(),
            })
    }

    pub fn should_be_multiline(&self) -> bool {
        self.has_tailing_comma_after_last_value()
            || self.has_multiline_values()
            || self.has_inner_comments()
    }

    pub fn has_tailing_comma_after_last_value(&self) -> bool {
        self.syntax()
            .children_with_tokens()
            .collect::<Vec<_>>()
            .into_iter()
            .rev()
            .skip_while(|item| item.kind() != T!(']'))
            .skip(1)
            .find(|item| !matches!(item.kind(), WHITESPACE | COMMENT | LINE_BREAK))
            .map_or(false, |it| it.kind() == T!(,))
    }

    pub fn has_multiline_values(&self) -> bool {
        self.values().any(|value| match value {
            crate::Value::Array(array) => array.should_be_multiline(),
            _ => false,
        })
    }

    pub fn has_inner_comments(&self) -> bool {
        support::has_inner_comments(self.syntax().children_with_tokens(), T!('['), T!(']'))
    }
}

impl InlineTable {
    pub fn inner_begin_dangling_comments(&self) -> impl Iterator<Item = crate::Comment> {
        support::begin_dangling_comments(
            self.syntax()
                .children_with_tokens()
                .skip_while(|node| node.kind() != T!('{'))
                .skip(1),
        )
    }

    pub fn inner_end_dangling_comments(&self) -> impl Iterator<Item = crate::Comment> {
        support::end_dangling_comments(
            self.syntax()
                .children_with_tokens()
                .take_while(|node| node.kind() != T!('}')),
        )
    }

    pub fn entries_with_comma(
        &self,
    ) -> impl Iterator<Item = (crate::KeyValue, Option<crate::Comma>)> {
        self.entries()
            .zip_longest(support::children::<crate::Comma>(self.syntax()))
            .map(|value_with_comma| match value_with_comma {
                itertools::EitherOrBoth::Both(value, comma) => (value, Some(comma)),
                itertools::EitherOrBoth::Left(value) => (value, None),
                itertools::EitherOrBoth::Right(_) => unreachable!(),
            })
    }

    pub fn should_be_multiline(&self, version: TomlVersion) -> bool {
        match version {
            TomlVersion::V1_0_0 => false,
            TomlVersion::V1_1_0_Preview => {
                self.has_multiline_values()
                    || self.has_tailing_comma_after_last_value()
                    || self.has_inner_comments()
            }
        }
    }

    pub fn has_tailing_comma_after_last_value(&self) -> bool {
        self.syntax()
            .children_with_tokens()
            .collect::<Vec<_>>()
            .into_iter()
            .rev()
            .skip_while(|item| item.kind() != T!('}'))
            .skip(1)
            .find(|item| !matches!(item.kind(), WHITESPACE | COMMENT | LINE_BREAK))
            .map_or(false, |it| it.kind() == T!(,))
    }

    pub fn has_multiline_values(&self) -> bool {
        self.entries().any(|entry| {
            entry.value().map_or(false, |value| match value {
                crate::Value::Array(array) => array.should_be_multiline(),
                _ => false,
            })
        })
    }

    pub fn has_inner_comments(&self) -> bool {
        support::has_inner_comments(self.syntax().children_with_tokens(), T!('{'), T!('}'))
    }
}
